import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spl
import scipy.linalg
sqrtm = scipy.linalg.sqrtm
norm = np.linalg.norm
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

eta = 0.001
mup = 0.001
mud = 0.0
N = 100

A = sp.csc_matrix(sp.diags([np.array([1] +  [1+eta] * (N-1)), -np.ones(N)], [0, -1]))
Q = mup * sp.eye(N).tocsc()
S = mud * sp.eye(N).tocsc()
u, s, v = spl.svds(A, 1)
NA = s[0]**2

b = np.array([1] + [0]*(N-1))
c = np.zeros(N)

gamma = 0.9
if mup * mud == 0:
    sigma = np.sqrt(gamma) / np.sqrt(NA)
    tau = gamma / (sigma * NA)
else:
    sigma = np.sqrt(mud / mup / NA)
    tau = np.sqrt(mup / mud / NA)

M1 = sp.csc_matrix(sp.vstack([sp.hstack([Q, A.T]),
                             sp.hstack([A, -S])]))

M = sp.csc_matrix(sp.vstack([sp.hstack([Q / tau, A.T / np.sqrt(sigma*tau)]),
                             sp.hstack([A / np.sqrt(sigma*tau), S / sigma])]))
u, s, v = spl.svds(M1, 1)
NM = s[0] ** 2

msr = min(np.abs(spl.eigs(M1, which = 'SM')[0]))
print('eta =', eta, 'mup =', mup, 'mud =', mud,  'msr =', msr)


def L(x, y):
    return 0.5 * x @ Q @ x + y @ (A @ x - b) - 0.5 * y @ S @ y


I = sp.eye(N)
R = sp.vstack([sp.hstack([(I - sigma * tau * A.T @ A) @ (sp.eye(N) - tau * Q),
                          - tau * A.T @ (I - sigma * tau * A @ A.T - sigma * S)]),
               sp.hstack([sigma * A @ (I - tau * Q),
                          I - sigma * tau * A @ A.T - sigma * S])]).tocsc()


def pdhg(x_init, y_init, tau, sigma, K):
    x = x_init.copy()
    y = y_init.copy()
    avgx = x.copy()
    avgy = y.copy()
    z_star = spl.spsolve(M1, np.hstack([-c, b]))
    x_star = z_star[:N]
    y_star = z_star[N:]

    Nx = np.zeros(K)
    Ny = np.zeros(K)
    tV = np.zeros((K,4))
    
    for k in range(K):
        x_ = spl.spsolve(I + tau * Q, x - tau * (A.T @ y + c))
        y_ = spl.spsolve(I + sigma * S, y + sigma * (A @ x_ - b))
        tV[k,0] = -tau/2. * norm(A.T @ (y_ - y))**2 + 0.5 * norm(x_ - x)**2/tau\
                + 0.5 * norm(y_ - y)**2/sigma
        tV[k,1] = L(x_, y_star) - L (x_star, y_)
        x_gamma = (x_star - tau * A.T @ y_) / (1+mup * tau)
        tV[k,2] = L(x_, y_star) + sigma / 2 * norm(A @ x_ - b)**2 \
                  + b.dot(y) - y @ A @ x_gamma - 0.5 * mup * norm(x_gamma)**2 \
                  - 0.5 / tau * norm(x_gamma - x_star)**2
        tV[k,3] = tau / 2 * norm(x_gamma - x_star)**2
        
        x = x_ - tau * A.T @ (y_ - y)
        y = y_.copy()
        avgx = k/(k+1) * avgx + 1./(k+1)*x_
        avgy = k/(k+1) * avgy + 1./(k+1)*y_
    
        Nx[k] = norm(x - x_star)
        Ny[k] = norm(y - y_star)

    return x, y, avgx, avgy, Nx, Ny, tV


def rapdhg(x_init, y_init, tau, sigma, period, K):
    x = x_init.copy()
    y = y_init.copy()

    x_star = spl.spsolve(A, b)
    y_star = spl.spsolve(A.T, -c)
    
    Nx = np.zeros(K)
    Ny = np.zeros(K)
    ktot = 0

    while ktot < K:
        x_in, y_in, avgx, avgy, Nx_in, Ny_in, _ = pdhg(x, y,  tau, sigma,
                                                    min(period, K-ktot))
        x = avgx
        y = avgy
        # x = x - tau * period * (A.T @ avgy + c)  # stable mais lent
        # y = y + sigma * period * (A @ avgx - b)
        Nx[ktot:ktot+len(Nx_in)] = Nx_in
        Ny[ktot:ktot+len(Ny_in)] = Ny_in
        ktot = ktot + period

    return x, y, x_in, y_in, Nx, Ny

if 0:
    K = 10000
    period = int(np.ceil(np.exp(1) * NM / msr))
    x, y, avgx, avgy, Nx, Ny, tV = pdhg(np.zeros(N), np.zeros(N), tau, sigma, K)
    x2, y2, avgx, avgy, Nx2, Ny2 = rapdhg(np.zeros(N), np.zeros(N), tau, sigma, period, K)
    V = Nx**2 / (2*tau) + Ny**2 / (2*sigma)

    mu_x = mup
    eta_x = msr
    L_x = mup
    L_y = max(1e-30, mud)
    mu_y = mud
    eta_y = min(spl.svds(A, which = 'SM')[1])
    #rho = min(max(mu_x * tau/2, eta_x ** 2 * min(tau / 2 / L_x, sigma * tau / (1 + sigma * L_y)**2)),
    #          max(mu_y * sigma / 2 , eta_y ** 2 * min(sigma / 2 / L_y, sigma * tau / (1+tau * L_x)**2)))
    beta = 1/(1-np.sqrt(gamma))
    lamb = min(1, beta * (1-np.sqrt(gamma)))
    rho = eta_y**2 / NM / beta

    plt.figure(1)
    plt.semilogy(np.arange(K), Nx ** 2 / tau + Ny ** 2 / sigma,
             np.arange(K), Nx2 ** 2 / tau + Ny2 ** 2 / sigma,
             np.arange(K), V[0] * (1 - (1-gamma) * (msr/NM)**2) ** np.arange(K),
             np.arange(K), V[0] * (1 - msr/NM) ** np.arange(K),
             np.arange(K), V[0] / (1 + lamb * rho / (1 + 2*rho)) ** np.arange(K)
    )

    plt.legend(['pdhg', 'rapdhg', 'slow_rate', 'fast_rate', 'slow_rate_2'])

    plt.show()
    


if 1:
    tau = 1.
    sigma = 1.
    grid = 100
    Result = []
    max_cond = 0.1

    a = 0.03
    Allqs = np.concatenate([np.arange(0, max_cond/tau/grid, max_cond/tau/grid/100), np.arange(max_cond/tau/grid, max_cond/tau, max_cond/tau/grid)])
    if 1: # for a in np.arange(max_cond/np.sqrt(sigma*tau)/grid, max_cond/np.sqrt(sigma*tau), max_cond/np.sqrt(sigma*tau)/grid):
        for q in Allqs:
            s_max = min([max_cond/sigma, tau/sigma*q*(1-tau*sigma*a**2)])
            s=0
            if 1:  #for s in np.arange(0, s_max, max_cond/sigma/grid):
                Rb = np.array([[(1-tau*q), -tau*a],
                              [sigma * a *(1-tau*q),
                                                1 - sigma*tau*a**2 -sigma*s]])
                R = np.array([[(1 - sigma * tau * a**2)*(1-tau*q),
                                                -tau*a*(1-sigma*tau*a**2)],
                              [sigma * a *(1-tau*q),
                                                1 - sigma*tau*a**2 -sigma*s]])
                eigsR = np.linalg.eig(R)[0]
                rho = max(np.abs(eigsR)) ** 2
                comp = np.iscomplex(eigsR[0])
                Dmat = np.array([[1/tau, 0], [0, 1/sigma]])
                svdsR = np.linalg.svd(sqrtm(Dmat) @ R @ np.linalg.inv(sqrtm(Dmat)))[1]
                rho2 = max(svdsR) ** 2

                Vmat = np.array([[1/tau, a], [a, 1/sigma]])
                rho2V = max(np.linalg.svd(sqrtm(Vmat) @ R @ sqrtm(np.linalg.inv(Vmat)))[1])**2
                rho2P = rho2
                ll = 0
                for lll in np.arange(0, 5, 0.1):
                    Pmat = np.array([[1/tau, lll * a/max(1e-30, q)], [0, 1/sigma]])
                    rho2P_ = max(np.linalg.svd(sqrtm(Pmat) @ R @ sqrtm(np.linalg.inv(Pmat)))[1])**2
                    if rho2P_ < rho2P:
                        rho2P = rho2P_
                        ll = lll

                eta = 0.5 * (np.sqrt(q**2 * tau**2 + 4*sigma*tau*a**2) - q*tau)
                gamma = sigma*tau*a**2
                alphaf = tau*q/2.

                theory_msr = 1 - (1-alphaf)*(1-np.sqrt(gamma))*eta**2 / (np.sqrt(3)*eta + 2 + 2 * np.sqrt(3)*alphaf)**2

                alphag = 0
                lamb_nonnex = 1 - alphaf + (1-gamma)*alphaf/2 - np.sqrt((1-alphaf)**2 * gamma + ((1-gamma)*alphaf)**2)
                switch = min(1, 10*q*tau)
                betax = np.sqrt(gamma)*(1-switch)+switch/lamb_nonnex
                betay = 1 / (1/betax - lamb_nonnex)

                a2 = (q*tau - 1) / (1 - q*tau/2)
                Lamb = lamb_nonnex / max((1+a2)*lamb_nonnex + 1/betax,
                                         (2+a2)*lamb_nonnex + 1/betay)
                
                qeb = min(q*tau + gamma/betay, gamma / (betax + tau * q))
                theory_qebsm = 1./(1 + Lamb * qeb / (qeb/lamb_nonnex + 1))

                betay = 1e10
                etax = q*tau + gamma / betay
                etay2 = gamma / (q*tau)
                lamb3 = max([0,
                             2*(np.sqrt(gamma)/2+etay2)/(q**2*tau**2*(1-np.sqrt(gamma))*(1-q*tau/2)) - 1,
                             (gamma/q**2/tau**2*(1 + 2/np.sqrt(gamma) + 2*q**2*tau**2*(np.sqrt(gamma)/2+etay2))+1+1/2/betay)/(1-np.sqrt(gamma))/(1-q*tau/2) - 1])
                rate2 = max([(1+np.sqrt(gamma)/2)/(1+np.sqrt(gamma)/2+etay2), (1+lamb3)/(1+lamb3+etay2)])

                betax = max(q * tau, 1 - 0.5*q*tau/np.sqrt(gamma))
                betay = 1
                etax = q*tau + gamma / betay
                etay = gamma / (betax + q * tau)
                alpha1 = q/(q+lamb_nonnex/2*tau*sigma)
                alpha2 = min(1/4,np.sqrt(etax))
                rho3 = 1
                for C in np.arange(0.01, 0.9, 0.02):
                    #C = max(0.25, 1 - etax / np.sqrt(gamma))
                    lambda1_mix = (1/betax - 1/betay+(1-alpha2-C)*etax/(2*gamma*(1-alpha2))- C*etax*(1/alpha2))/(2*q*tau*(1-alpha1))
                    lambda1_alone = (1/betax + a2 * lamb_nonnex)/(2*q*tau*(1-alpha1)-C*etax)
                    if 2*q*tau*(1-alpha1) <= C*etax:
                        lambda1 = 0
                        lambda4 = (1/betax - lambda1 * (2*q*tau*(1-alpha1)-C*etax)) / lamb_nonnex + a2
                    elif lambda1_alone > lambda1_mix:
                        lambda1 = lambda1_mix
                        lambda4 = (1/betax - lambda1 * (2*q*tau*(1-alpha1)-C*etax)) / lamb_nonnex + a2
                    else:
                        lambda1 = lambda1_alone
                        lambda4 = 0
                    rho3_ = (1+lambda4) / (1 + lambda4 + min(etay, C*etax))
                    if rho3_ < rho3:
                        rho3 = rho3_
                        C_final = C
                print(q, rho3, C_final)
                
                qeb_strconv = min(q*tau, gamma/(tau * q + 1/lamb_nonnex))

                rho_strong_conv = 1./(1 + qeb_strconv / ((2+a2) * (1+qeb_strconv/lamb_nonnex)) )

                Result.append([a, q, s, rho, comp, theory_msr, theory_qebsm, Lamb, betax, qeb, rho2, rho2V, rho2P, rate2, rho3, rho_strong_conv])
    Result = np.array(Result)

    ill_cond_a = Result

    fig = plt.figure(1)
    ax = fig.add_subplot(111, projection='3d')
    h = ax.scatter(ill_cond_a[:,0], ill_cond_a[:,1], ill_cond_a[:,2], marker='o', c=ill_cond_a[:,4])
    plt.colorbar(h)
    ax.set_xlabel('a'); ax.set_ylabel('q'); ax.set_zlabel('s')
    # plt.show()


    ill_cond_s = Result[Result[:,2] == 0, :]
    fig = plt.figure(2)
    ax = fig.add_subplot(111, projection='3d')
    h = ax.scatter(ill_cond_s[:,0], ill_cond_s[:,1], ill_cond_s[:,3], marker='o', c=ill_cond_s[:,3])
    plt.colorbar(h)
    ax.set_xlabel('a'); ax.set_ylabel('q')
    # plt.show()

    fig = plt.figure(3)
    xaxis = Allqs
    plt.loglog(xaxis, 1-Result[:,3])
    plt.loglog(xaxis, 1-Result[:,15], linestyle='--')
    plt.loglog(xaxis, 1-Result[:,5], linestyle='-.')
    plt.loglog(xaxis, 1-Result[:,6], linestyle=':')
    plt.loglog(xaxis, 1-Result[:,14], linestyle=(0,(5,10)))
    plt.ylabel('1-rate'); plt.xlabel('µ')

    plt.legend(['true asymptotic rate',
                'theory using strong convexity',
                'theory using metric sub-regularity',
                'theory by quadratic error bound of the smoothed gap',
                'slow-fast double concentration rate'])

    fig = plt.figure(5)
    xaxis = Allqs
    plt.loglog(xaxis, 1-Result[:,3])
    plt.loglog(xaxis, 1-Result[:,15], linestyle='--')
    plt.loglog(xaxis, 1-Result[:,5], linestyle='-.')
    
    plt.legend(['true asymptotic rate',
                'theory using strong convexity',
                'theory using metric sub-regularity',
                ])
    plt.xlabel('µ')
    plt.ylabel('1 - rate')
    plt.show()



if 1:
    z_star=[0,0]
    x_star = 0
    y_star = 0
    x = 0
    y = 1
    xb = x - tau * (q*x + a*y)
    yb = y - sigma*(0 - a *xb)
    x_ = x1 = xb - tau*a*(yb-y)
    y_ = y1 = yb
    xb2 = x_ - tau * (q*x1 + a*y1)
    yb2 = y_ - sigma*(0 - a *xb2)
    x2 = xb2 - tau*a*(yb2-y1)
    y2 = yb2
    
    dF2 = q/2*xb**2 - (q/2*x**2+q*x*(xb-x) + q/2*(xb-x)**2)
    dG2 = 0.
    dF = (q*x+a*y)*(z_star[0]-xb) + 0.5/tau*x**2 - 0.5/tau*xb**2 - 0.5/tau*(xb-x)**2
    dG = b*yb - ((b*y - a* xb) * (z_star[1]-yb)+0.5/sigma*(y**2-yb**2-(y-yb)**2))

    LL = q*xb**2/2 + xb*a*z_star[1] - z_star[0]**2*q/2 - z_star[0]*a*yb
    DV = 0.5/tau *(x-z_star[0])**2 + 0.5/sigma*(y-z_star[1])**2 - 0.5/tau *(x_-z_star[0])**2 - 0.5/sigma*(y_-z_star[1])**2
    qV = (0.5/tau *(x-z_star[0])**2 + 0.5/sigma*(y-z_star[1])**2) / (0.5/tau *(x_-z_star[0])**2 + 0.5/sigma*(y_-z_star[1])**2)
    tV = (0.5/tau-q/2)*(xb-x)**2 +(0.5/sigma - tau*a**2/2)*(yb-y)**2
    strconv_bonus = 0.5*(z_star[0] - x)**2 * q

    def L(z):
        x = z[0]
        y = z[1]
        return q*x**2/2 + (a*x-b)*y

    def dL(z1, z2):
        return L([z1[0], z2[1]]) - L([z2[0], z1[1]])

    def V(z):
        x = z[0]
        y = z[1]
        return 0.5/tau * (x)**2 + 0.5/sigma * (y)**2

    def dV(z, z_ref):
        z = np.array(z)
        z_ref = np.array(z_ref)
        return V(z - z_ref) - V(R @ z - z_ref)

    def tildeV(z):
        zb = Rb @ z
        return (0.5/tau-q/2)*(zb[0]-z[0])**2 +(0.5/sigma - tau*a**2/2)*(zb[1]-z[1])**2

    def diff(z, z_ref):
        return dL(z, z_ref) - dV(z, z_ref) - tildeV(z)

    def smoothed_gap(beta, z, z_ref):
        if np.isscalar(beta) == True:
            beta = [beta, beta]
        betax = beta[0]
        betay = beta[1]
        zb = Rb @ z
        xb = zb[0]
        yb = zb[1]
        x_star = z_ref[0]
        y_star = z_ref[1]
        xx_lhs = (betax / tau * x_star - a * yb) / (betax/tau + q)
        yy_lhs = y_star + sigma/betay * (a * xb - b)
        return q*xb**2/2 + xb*a*yy_lhs - yy_lhs*b - 0.5*betay/sigma*(y_star-yy_lhs)**2 - xx_lhs**2*q/2 - xx_lhs*a*yb + yb*b - 0.5*betax/tau*(x_star-xx_lhs)**2

    
if 1:
    q = 0.1
    b = 0
    RR = []
    for betax in [max([0.1,(gamma - (q*tau)**2)/(q*tau)])]: # 1000 * 0.5**np.arange(20):
        for betay in [1e10]: #1000 * 0.5**np.arange(20):
            gamma = sigma*tau*a**2
            lamb = (1 - q*tau/2) * (1-np.sqrt(gamma)) / max([1/betax, 1+1/betay])
            etax = q*tau + gamma/betay
            etay = gamma/(betax + q*tau)
            eta = min([etax, etay])
            rate = 1. / (1+lamb * eta / (1+eta))

            etay2 = gamma / (q*tau)
            lamb3 = max([0, 2*(gamma+2*etay2)/(q**2*tau**2*(1-np.sqrt(gamma))*(1-q*tau/2)) - 1])
            rate2 = max([(1+gamma)/(1+gamma+2*etay2), lamb3 / (lamb3 + etax/(1+etax)), (1+lamb3)/(1+lamb3+etay2)])

            RR.append([betax, betay, rate])

            xx_lhs = (betax / tau * x_star - a * yb) / (betax/tau + q)
            yy_lhs = y_star + sigma/betay * (a * xb - b)
            xx_rhs = (betax * x_star + x_ - (1-q*tau) * x) / (betax + q*tau)
            yy_rhs = y_star + (y_ - y) / betay

            Gb = q*xb**2/2 + xb*a*yy_lhs - yy_lhs*b - 0.5*betay/sigma*(y_star-yy_lhs)**2 - xx_lhs**2*q/2 - xx_lhs*a*yb + yb*b - 0.5*betax/tau*(x_star-xx_lhs)**2
            rhs = 0.5*(1/tau-q) *(x-xx_rhs)**2 + 0.5/sigma*(y-yy_rhs)**2 - 0.5/tau *(x_-xx_rhs)**2 - 0.5/sigma*(y_-yy_rhs)**2 -tV -0.5*((xx_rhs-x_star)**2*betax/tau + ((yy_rhs-y_star)**2)*betay/sigma)
            DV2 = 0.5*(1/tau-q/(betax+tau*q))*(x-z_star[0])**2 + 0.5/sigma*(y-z_star[1])**2 - 0.5*(1/tau-q/(betax+tau*q))*(x_-z_star[0])**2 - 0.5/sigma*(y_-z_star[1])**2
            additional_term = 0.5*(xb-x)**2*(1/tau-q)/(betax+q*tau) + 0.5*(yb-y)**2/sigma/betay
            bonus_term = 0.5*q*(x-z_star[0])**2 - 0.5*(q**2*tau/(betax+tau*q))*(x-z_star[0])**2

    # new rate with z_k-z^* and z_k-z_{k-1}
    betax = max(q*tau, 1 - q*tau / np.sqrt(gamma))
    betay = 1
    etax = q*tau + gamma / betay
    etay = gamma / (betax + q * tau)
    rho3 = 1 / min(1 + etax / 4, 1 + etay,
                   1 + etax * q * tau / 4 * (etax * (1/4/gamma - 0.5) + (1-np.sqrt(gamma))*(1-alphaf)) / (1 - q * tau * ((1-np.sqrt(gamma))*(1-alphaf))))
            
    RR = np.array(RR)
    # plt.close('all')

    upper_bound_0 = 0.5/(betax)/tau*(x_ - x)**2
    upper_bound_mu = 0.5/(betax+q*tau)/tau*(x_ - x + q*tau*(x-x_star))**2

    lambda3=1
    lambda1max = q**2*tau**2/gamma*((1+lambda3)*(1-np.sqrt(gamma)) -1)/(1+gamma+2*q**2*tau**2*(gamma+2*eta))
    tmp = (gamma+2*eta)/tau**2/q**2
    lambda2min = betax*( tmp - np.sqrt(tmp**2 +2/betax*(-tmp+0.5*(1+lambda3)*(1-np.sqrt(gamma)))))
    lambda2max = betax*( tmp + np.sqrt(tmp**2 +2/betax*(-tmp+0.5*(1+lambda3)*(1-np.sqrt(gamma)))))

    print(lambda3, lambda1max, 1-lambda2min, lambda2max)
    
    
